@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2010, 2011, 2012, 2013, 2014, 2015 Nicira, Inc.
+ * Copyright (c) 2010, 2011, 2012, 2013, 2014, 2015, 2016 Nicira, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -16,29 +16,29 @@
 
 #include <config.h>
 
-#include "ofp-parse.h"
-
 #include <ctype.h>
 #include <errno.h>
 #include <stdlib.h>
 #include <netinet/in.h>
 
 #include "byte-order.h"
-#include "dynamic-string.h"
 #include "learn.h"
-#include "meta-flow.h"
 #include "multipath.h"
 #include "netdev.h"
 #include "nx-match.h"
-#include "ofp-actions.h"
-#include "ofp-util.h"
-#include "ofpbuf.h"
 #include "openflow/openflow.h"
+#include "openvswitch/dynamic-string.h"
+#include "openvswitch/meta-flow.h"
+#include "openvswitch/ofp-actions.h"
+#include "openvswitch/ofp-parse.h"
+#include "openvswitch/ofp-util.h"
+#include "openvswitch/ofpbuf.h"
+#include "openvswitch/vconn.h"
 #include "ovs-thread.h"
 #include "packets.h"
 #include "simap.h"
 #include "socket-util.h"
-#include "openvswitch/vconn.h"
+#include "util.h"
 
 /* Parses 'str' as an 8-bit unsigned integer into '*valuep'.
  *
@@ -332,27 +332,20 @@ parse_ofp_str__(struct ofputil_flow_mod *fm, int command, char *string,
         OVS_NOT_REACHED();
     }
 
-    match_init_catchall(&fm->match);
-    fm->priority = OFP_DEFAULT_PRIORITY;
-    fm->cookie = htonll(0);
-    fm->cookie_mask = htonll(0);
+    *fm = (struct ofputil_flow_mod) {
+        .match = MATCH_CATCHALL_INITIALIZER,
+        .priority = OFP_DEFAULT_PRIORITY,
+        .table_id = 0xff,
+        .command = command,
+        .buffer_id = UINT32_MAX,
+        .out_port = OFPP_ANY,
+        .out_group = OFPG_ANY,
+    };
+    /* For modify, by default, don't update the cookie. */
     if (command == OFPFC_MODIFY || command == OFPFC_MODIFY_STRICT) {
-        /* For modify, by default, don't update the cookie. */
         fm->new_cookie = OVS_BE64_MAX;
-    } else{
-        fm->new_cookie = htonll(0);
     }
-    fm->modify_cookie = false;
-    fm->table_id = 0xff;
-    fm->command = command;
-    fm->idle_timeout = OFP_FLOW_PERMANENT;
-    fm->hard_timeout = OFP_FLOW_PERMANENT;
-    fm->buffer_id = UINT32_MAX;
-    fm->out_port = OFPP_ANY;
-    fm->flags = 0;
-    fm->importance = 0;
-    fm->out_group = OFPG_ANY;
-    fm->delete_reason = OFPRR_DELETE;
+
     if (fields & F_ACTIONS) {
         act_str = extract_actions(string);
         if (!act_str) {
@@ -1121,23 +1114,23 @@ parse_ofp_flow_stats_request_str(struct ofputil_flow_stats_request *fsr,
 /* Parses a specification of a flow from 's' into 'flow'.  's' must take the
  * form FIELD=VALUE[,FIELD=VALUE]... where each FIELD is the name of a
  * mf_field.  Fields must be specified in a natural order for satisfying
- * prerequisites. If 'mask' is specified, fills the mask field for each of the
+ * prerequisites. If 'wc' is specified, masks the field in 'wc' for each of the
  * field specified in flow. If the map, 'names_portno' is specfied, converts
  * the in_port name into port no while setting the 'flow'.
  *
  * Returns NULL on success, otherwise a malloc()'d string that explains the
  * problem. */
 char *
-parse_ofp_exact_flow(struct flow *flow, struct flow *mask, const char *s,
-                     const struct simap *portno_names)
+parse_ofp_exact_flow(struct flow *flow, struct flow_wildcards *wc,
+                     const char *s, const struct simap *portno_names)
 {
     char *pos, *key, *value_s;
     char *error = NULL;
     char *copy;
 
     memset(flow, 0, sizeof *flow);
-    if (mask) {
-        memset(mask, 0, sizeof *mask);
+    if (wc) {
+        memset(wc, 0, sizeof *wc);
     }
 
     pos = copy = xstrdup(s);
@@ -1149,8 +1142,8 @@ parse_ofp_exact_flow(struct flow *flow, struct flow *mask, const char *s,
                 goto exit;
             }
             flow->dl_type = htons(p->dl_type);
-            if (mask) {
-                mask->dl_type = OVS_BE16_MAX;
+            if (wc) {
+                wc->masks.dl_type = OVS_BE16_MAX;
             }
 
             if (p->nw_proto) {
@@ -1160,8 +1153,8 @@ parse_ofp_exact_flow(struct flow *flow, struct flow *mask, const char *s,
                     goto exit;
                 }
                 flow->nw_proto = p->nw_proto;
-                if (mask) {
-                    mask->nw_proto = UINT8_MAX;
+                if (wc) {
+                    wc->masks.nw_proto = UINT8_MAX;
                 }
             }
         } else {
@@ -1175,7 +1168,7 @@ parse_ofp_exact_flow(struct flow *flow, struct flow *mask, const char *s,
                 goto exit;
             }
 
-            if (!mf_are_prereqs_ok(mf, flow)) {
+            if (!mf_are_prereqs_ok(mf, flow, NULL)) {
                 error = xasprintf("%s: prerequisites not met for setting %s",
                                   s, key);
                 goto exit;
@@ -1191,8 +1184,9 @@ parse_ofp_exact_flow(struct flow *flow, struct flow *mask, const char *s,
                 && simap_contains(portno_names, value_s)) {
                 flow->in_port.ofp_port = u16_to_ofp(
                     simap_get(portno_names, value_s));
-                if (mask) {
-                    mask->in_port.ofp_port = u16_to_ofp(ntohs(OVS_BE16_MAX));
+                if (wc) {
+                    wc->masks.in_port.ofp_port
+                        = u16_to_ofp(ntohs(OVS_BE16_MAX));
                 }
             } else {
                 field_error = mf_parse_value(mf, value_s, &value);
@@ -1204,8 +1198,8 @@ parse_ofp_exact_flow(struct flow *flow, struct flow *mask, const char *s,
                 }
 
                 mf_set_flow_value(mf, &value, flow);
-                if (mask) {
-                    mf_mask_field(mf, mask);
+                if (wc) {
+                    mf_mask_field(mf, wc);
                 }
             }
         }
@@ -1220,8 +1214,8 @@ exit:
 
     if (error) {
         memset(flow, 0, sizeof *flow);
-        if (mask) {
-            memset(mask, 0, sizeof *mask);
+        if (wc) {
+            memset(wc, 0, sizeof *wc);
         }
     }
     return error;
@@ -1357,7 +1351,7 @@ parse_select_group_field(char *s, struct field_array *fa,
 }
 
 static char * OVS_WARN_UNUSED_RESULT
-parse_ofp_group_mod_str__(struct ofputil_group_mod *gm, uint16_t command,
+parse_ofp_group_mod_str__(struct ofputil_group_mod *gm, int command,
                           char *string,
                           enum ofputil_protocol *usable_protocols)
 {
@@ -1374,6 +1368,31 @@ parse_ofp_group_mod_str__(struct ofputil_group_mod *gm, uint16_t command,
 
     *usable_protocols = OFPUTIL_P_OF11_UP;
 
+    if (command == -2) {
+        size_t len;
+
+        string += strspn(string, " \t\r\n");   /* Skip white space. */
+        len = strcspn(string, ", \t\r\n"); /* Get length of the first token. */
+
+        if (!strncmp(string, "add", len)) {
+            command = OFPGC11_ADD;
+        } else if (!strncmp(string, "delete", len)) {
+            command = OFPGC11_DELETE;
+        } else if (!strncmp(string, "modify", len)) {
+            command = OFPGC11_MODIFY;
+        } else if (!strncmp(string, "add_or_mod", len)) {
+            command = OFPGC11_ADD_OR_MOD;
+        } else if (!strncmp(string, "insert_bucket", len)) {
+            command = OFPGC15_INSERT_BUCKET;
+        } else if (!strncmp(string, "remove_bucket", len)) {
+            command = OFPGC15_REMOVE_BUCKET;
+        } else {
+            len = 0;
+            command = OFPGC11_ADD;
+        }
+        string += len;
+    }
+
     switch (command) {
     case OFPGC11_ADD:
         fields = F_GROUP_TYPE | F_BUCKETS;
@@ -1387,6 +1406,10 @@ parse_ofp_group_mod_str__(struct ofputil_group_mod *gm, uint16_t command,
         fields = F_GROUP_TYPE | F_BUCKETS;
         break;
 
+    case OFPGC11_ADD_OR_MOD:
+        fields = F_GROUP_TYPE | F_BUCKETS;
+        break;
+
     case OFPGC15_INSERT_BUCKET:
         fields = F_BUCKETS | F_COMMAND_BUCKET_ID;
         *usable_protocols &= OFPUTIL_P_OF15_UP;
@@ -1405,7 +1428,7 @@ parse_ofp_group_mod_str__(struct ofputil_group_mod *gm, uint16_t command,
     gm->command = command;
     gm->group_id = OFPG_ANY;
     gm->command_bucket_id = OFPG15_BUCKET_ALL;
-    list_init(&gm->buckets);
+    ovs_list_init(&gm->buckets);
     if (command == OFPGC11_DELETE && string[0] == '\0') {
         gm->group_id = OFPG_ALL;
         return NULL;
@@ -1543,6 +1566,18 @@ parse_ofp_group_mod_str__(struct ofputil_group_mod *gm, uint16_t command,
         goto out;
     }
 
+    /* Exclude fields for non "hash" selection method. */
+    if (strcmp(gm->props.selection_method, "hash") &&
+        gm->props.fields.values_size) {
+        error = xstrdup("fields may only be specified with \"selection_method=hash\"");
+        goto out;
+    }
+    /* Exclude selection_method_param if no selection_method is given. */
+    if (gm->props.selection_method[0] == 0
+        && gm->props.selection_method_param != 0) {
+        error = xstrdup("selection_method_param is only allowed with \"selection_method\"");
+        goto out;
+    }
     if (fields & F_COMMAND_BUCKET_ID) {
         if (!(fields & F_COMMAND_BUCKET_ID_ALL || had_command_bucket_id)) {
             error = xstrdup("must specify a command bucket id");
@@ -1575,7 +1610,7 @@ parse_ofp_group_mod_str__(struct ofputil_group_mod *gm, uint16_t command,
             free(bucket);
             goto out;
         }
-        list_push_back(&gm->buckets, &bucket->list_node);
+        ovs_list_push_back(&gm->buckets, &bucket->list_node);
 
         if (gm->type != OFPGT11_SELECT && bucket->weight) {
             error = xstrdup("Only select groups can have bucket weights.");
@@ -1584,19 +1619,23 @@ parse_ofp_group_mod_str__(struct ofputil_group_mod *gm, uint16_t command,
 
         bkt_str = next_bkt_str;
     }
-    if (gm->type == OFPGT11_INDIRECT && !list_is_short(&gm->buckets)) {
+    if (gm->type == OFPGT11_INDIRECT && !ovs_list_is_short(&gm->buckets)) {
         error = xstrdup("Indirect groups can have at most one bucket.");
         goto out;
     }
 
     return NULL;
  out:
-    ofputil_bucket_list_destroy(&gm->buckets);
+    ofputil_uninit_group_mod(gm);
     return error;
 }
 
+/* If 'command' is given as -2, each line may start with a command name ("add",
+ * "modify", "add_or_mod", "delete", "insert_bucket", or "remove_bucket").  A
+ * missing command name is treated as "add".
+ */
 char * OVS_WARN_UNUSED_RESULT
-parse_ofp_group_mod_str(struct ofputil_group_mod *gm, uint16_t command,
+parse_ofp_group_mod_str(struct ofputil_group_mod *gm, int command,
                         const char *str_,
                         enum ofputil_protocol *usable_protocols)
 {
@@ -1604,15 +1643,15 @@ parse_ofp_group_mod_str(struct ofputil_group_mod *gm, uint16_t command,
     char *error = parse_ofp_group_mod_str__(gm, command, string,
                                             usable_protocols);
     free(string);
-
-    if (error) {
-        ofputil_bucket_list_destroy(&gm->buckets);
-    }
     return error;
 }
 
+/* If 'command' is given as -2, each line may start with a command name ("add",
+ * "modify", "add_or_mod", "delete", "insert_bucket", or "remove_bucket").  A
+ * missing command name is treated as "add".
+ */
 char * OVS_WARN_UNUSED_RESULT
-parse_ofp_group_mod_file(const char *file_name, uint16_t command,
+parse_ofp_group_mod_file(const char *file_name, int command,
                          struct ofputil_group_mod **gms, size_t *n_gms,
                          enum ofputil_protocol *usable_protocols)
 {
@@ -1644,7 +1683,7 @@ parse_ofp_group_mod_file(const char *file_name, uint16_t command,
 
             new_gms = x2nrealloc(*gms, &allocated_gms, sizeof **gms);
             for (i = 0; i < *n_gms; i++) {
-                list_moved(&new_gms[i].buckets, &(*gms)[i].buckets);
+                ovs_list_moved(&new_gms[i].buckets, &(*gms)[i].buckets);
             }
             *gms = new_gms;
         }
@@ -1654,7 +1693,7 @@ parse_ofp_group_mod_file(const char *file_name, uint16_t command,
             size_t i;
 
             for (i = 0; i < *n_gms; i++) {
-                ofputil_bucket_list_destroy(&(*gms)[i].buckets);
+                ofputil_uninit_group_mod(&(*gms)[i]);
             }
             free(*gms);
             *gms = NULL;
@@ -1665,7 +1704,9 @@ parse_ofp_group_mod_file(const char *file_name, uint16_t command,
                 fclose(stream);
             }
 
-            return xasprintf("%s:%d: %s", file_name, line_number, error);
+            char *ret = xasprintf("%s:%d: %s", file_name, line_number, error);
+            free(error);
+            return ret;
         }
         *usable_protocols &= usable;
         *n_gms += 1;
@@ -1678,6 +1719,119 @@ parse_ofp_group_mod_file(const char *file_name, uint16_t command,
     return NULL;
 }
 
+static void
+free_bundle_msgs(struct ofputil_bundle_msg **bms, size_t *n_bms)
+{
+    for (size_t i = 0; i < *n_bms; i++) {
+        switch ((int)(*bms)[i].type) {
+        case OFPTYPE_FLOW_MOD:
+            free(CONST_CAST(struct ofpact *, (*bms)[i].fm.ofpacts));
+            break;
+        case OFPTYPE_GROUP_MOD:
+            ofputil_uninit_group_mod(&(*bms)[i].gm);
+            break;
+        default:
+            break;
+        }
+    }
+    free(*bms);
+    *bms = NULL;
+    *n_bms = 0;
+}
+
+/* Opens file 'file_name' and reads each line as a flow_mod or a group_mod,
+ * depending on the first keyword on each line.  Stores each flow and group
+ * mods in '*bms', an array allocated on the caller's behalf, and the number of
+ * messages in '*n_bms'.
+ *
+ * Returns NULL if successful, otherwise a malloc()'d string describing the
+ * error.  The caller is responsible for freeing the returned string. */
+char * OVS_WARN_UNUSED_RESULT
+parse_ofp_bundle_file(const char *file_name,
+                      struct ofputil_bundle_msg **bms, size_t *n_bms,
+                      enum ofputil_protocol *usable_protocols)
+{
+    size_t allocated_bms;
+    char *error = NULL;
+    int line_number;
+    FILE *stream;
+    struct ds ds;
+
+    *usable_protocols = OFPUTIL_P_ANY;
+
+    *bms = NULL;
+    *n_bms = 0;
+
+    stream = !strcmp(file_name, "-") ? stdin : fopen(file_name, "r");
+    if (stream == NULL) {
+        return xasprintf("%s: open failed (%s)",
+                         file_name, ovs_strerror(errno));
+    }
+
+    allocated_bms = *n_bms;
+    ds_init(&ds);
+    line_number = 0;
+    while (!ds_get_preprocessed_line(&ds, stream, &line_number)) {
+        enum ofputil_protocol usable;
+        char *s = ds_cstr(&ds);
+        size_t len;
+
+        if (*n_bms >= allocated_bms) {
+            struct ofputil_bundle_msg *new_bms;
+
+            new_bms = x2nrealloc(*bms, &allocated_bms, sizeof **bms);
+            for (size_t i = 0; i < *n_bms; i++) {
+                if (new_bms[i].type == OFPTYPE_GROUP_MOD) {
+                    ovs_list_moved(&new_bms[i].gm.buckets,
+                                   &(*bms)[i].gm.buckets);
+                }
+            }
+            *bms = new_bms;
+        }
+
+        s += strspn(s, " \t\r\n");   /* Skip white space. */
+        len = strcspn(s, ", \t\r\n"); /* Get length of the first token. */
+
+        if (!strncmp(s, "flow", len)) {
+            s += len;
+            error = parse_ofp_flow_mod_str(&(*bms)[*n_bms].fm, s, -2, &usable);
+            if (error) {
+                break;
+            }
+            (*bms)[*n_bms].type = OFPTYPE_FLOW_MOD;
+        } else if (!strncmp(s, "group", len)) {
+            s += len;
+            error = parse_ofp_group_mod_str(&(*bms)[*n_bms].gm, -2, s,
+                                            &usable);
+            if (error) {
+                break;
+            }
+            (*bms)[*n_bms].type = OFPTYPE_GROUP_MOD;
+        } else {
+            error = xasprintf("Unsupported bundle message type: %.*s",
+                              (int)len, s);
+            break;
+        }
+
+        *usable_protocols &= usable; /* Each line can narrow the set. */
+        *n_bms += 1;
+    }
+
+    ds_destroy(&ds);
+    if (stream != stdin) {
+        fclose(stream);
+    }
+
+    if (error) {
+        char *err_msg = xasprintf("%s:%d: %s", file_name, line_number, error);
+        free(error);
+
+        free_bundle_msgs(bms, n_bms);
+        return err_msg;
+    }
+    return NULL;
+}
+
 char * OVS_WARN_UNUSED_RESULT
 parse_ofp_tlv_table_mod_str(struct ofputil_tlv_table_mod *ttm,
                                uint16_t command, const char *s,
@@ -1686,7 +1840,7 @@ parse_ofp_tlv_table_mod_str(struct ofputil_tlv_table_mod *ttm,
     *usable_protocols = OFPUTIL_P_NXM_OXM_ANY;
 
     ttm->command = command;
-    list_init(&ttm->mappings);
+    ovs_list_init(&ttm->mappings);
 
     while (*s) {
         struct ofputil_tlv_map *map = xmalloc(sizeof *map);
@@ -1696,7 +1850,7 @@ parse_ofp_tlv_table_mod_str(struct ofputil_tlv_table_mod *ttm,
             s++;
         }
 
-        list_push_back(&ttm->mappings, &map->list_node);
+        ovs_list_push_back(&ttm->mappings, &map->list_node);
 
         if (!ovs_scan(s, "{class=%"SCNi16",type=%"SCNi8",len=%"SCNi8"}->tun_metadata%"SCNi16"%n",
                       &map->option_class, &map->option_type, &map->option_len,
