@@ -844,6 +844,7 @@ parse_ct_lb_action(struct action_context *ctx)
         while (!lexer_match(ctx->lexer, LEX_T_RPAREN)) {
             if (ctx->lexer->token.type != LEX_T_INTEGER
                 || mf_subvalue_width(&ctx->lexer->token.value) > 32) {
+                free(dsts);
                 lexer_syntax_error(ctx->lexer, "expecting IPv4 address");
                 return;
             }
@@ -1527,21 +1528,26 @@ encode_put_dhcpv6_option(const struct ovnact_dhcp_option *o,
                          struct ofpbuf *ofpacts)
 {
     struct dhcp_opt6_header *opt = ofpbuf_put_uninit(ofpacts, sizeof *opt);
-    opt->code = o->option->code;
-
     const union expr_constant *c = o->value.values;
     size_t n_values = o->value.n_values;
+    size_t size;
+
+    opt->opt_code = htons(o->option->code);
+
     if (!strcmp(o->option->type, "ipv6")) {
-        opt->len = n_values * sizeof(struct in6_addr);
+        size = n_values * sizeof(struct in6_addr);
+        opt->size = htons(size);
         for (size_t i = 0; i < n_values; i++) {
             ofpbuf_put(ofpacts, &c[i].value.ipv6, sizeof(struct in6_addr));
         }
     } else if (!strcmp(o->option->type, "mac")) {
-        opt->len = sizeof(struct eth_addr);
-        ofpbuf_put(ofpacts, &c->value.mac, opt->len);
+        size = sizeof(struct eth_addr);
+        opt->size = htons(size);
+        ofpbuf_put(ofpacts, &c->value.mac, size);
     } else if (!strcmp(o->option->type, "str")) {
-        opt->len = strlen(c->string);
-        ofpbuf_put(ofpacts, c->string, opt->len);
+        size = strlen(c->string);
+        opt->size = htons(size);
+        ofpbuf_put(ofpacts, c->string, size);
     }
 }
 
